* SUBROUTINE UXSSCG                ALL SYSTEMS                06/12/01
C PORTABILITY : ALL SYSTEMS
C 06/12/01 MA : ORIGINAL VERSION
*
* PURPOSE :
* PRECONDITIONED CONJUGATE GRADIENT METHOD FOR THE SYSTEM OF LINEAR EQUATIONS
* WITH A SPARSE SYMMETRIC MATRIX AUGMENTED BY ONE ROW AND ONE COLUMN.
* THE SYSTEM HAS THE FOLLOWING FORM:
*
*  |    G      A  | . | X  | = |   Z   |
*  | TRANS(A)  MY |   | XI |   | OMEGA |
*
* PARAMETERS :
*  II  N  ORDER OF THE SPARSE SYMMETRIC MATRIX G.
*  II  M  NUMBER OF NONZERO ELEMENTS OF THE MATRIX G.
*  RI  G(M) ELEMENTS OF THE SPARSE SYMMETRIC MATRIX STORED IN THE
*      PACKED FORM.
*  II  IG(N+1) POINTERS OF THE MATRIX G STORED IN A COMPACT FORM.
*  II  JG(M) COLUMN INDICES OF NONZEROES IN THE MATRIX G.
*  RI  A(N)  VECTOR AUGMENTING THE SYSTEM.
*  RI  MY  THE LAST DIAGONAL ELEMENT OF THE AUGMENTED SYSTEM.
*  RI  Z(N)  THE RIGHT HAND SIDE.
*  RI  OMEGA  THE LAST ELEMENT OF THE RIGHT HAND SIDE.
*  RO  X(N) SOLUTION OF THE SYSTEM.
*  RO  XI  THE LAST ELEMENT OF THE SOLUTION OF THE SYSTEM.
*  RA  WD(N) AUXILIARY VECTOR.
*  RA  WA(N) AUXILIARY VECTOR.
*  RA  WR(N) AUXILIARY VECTOR.
*  RA  WT(N) AUXILIARY VECTOR.
*  RA  WP(N) AUXILIARY VECTOR.
*  RA  WQ(N) AUXILIARY VECTOR.
*  IA  IW(N) AUXILIARY VECTOR.
*  RI  EPS6  STOPPING TOLERANCE FOR THE SOLUTION.
*  RI  ETA2  A DESIRED TOLERANCE FOR POSITIVE DEFINITENESS FOR THE
*      INCOMPLETE CHOLESKY FACTORIZATION.
*  II  MOS2  TYPE OF PRECONDITIONING: 0 - NO PRECONDITIONING,
*      1 - DIAGONAL, 2 - INCOMPLETE CHOLESKY FACTORIZATION, 3 - SSOR.
*  II  MOS4
*  IU  NDEC  NUMBER OF MATRIX DECOMPOSITIONS.
*  IU  NMV  NUMBER OF MATRIX-VECTOR MULTIPLICATIONS.
*  II  MCG  MAXIMUM NUMBER OF CONJUGATE GRADIENT ITERATIONS.
*  IO  INF  INFORMATION ABOUT TERMINATION:
*         INF = I   SUCCESSFUL TERMINATION IN THE I-TH ITERATION,
*         INF = 0   THE LAST DIAGONAL ELEMENT OF THE AUGMENTED SYSTEM
*                   EQUALS PRECONDITIONING DOT PRODUCT,
*         INF = -I  NEGATIVE CURVATURE IS FOUND IN THE I-TH ITERATION.
*
* METHOD :
*  LUKSAN L., MATONOHA C., VLCEK J.: INTERIOR POINT METHOD FOR NON-LINEAR
*  NON-CONVEX OPTIMIZATION, NUMERICAL LINEAR ALGEBRA WITH APPLICATIONS,
*  VOL. 11, NO. 5-6, 2004, PP. 431-453.
*
* SUBPROGRAMS USED :
*
*  S   UXSSDC  DIAGONAL OF A SPARSE SYMMETRIC MATRIX IS COPIED TO THE VECTOR.
*  S   UXVMUL  VECTOR IS PREMULTIPLIED BY THE POWER OF A DIAGONAL MATRIX.
*  S   UXVCOP  COPYING OF A VECTOR.
*  S   UXSPTF  INCOMPLETE CHOLESKY FACTORIZATION OF A MATRIX.
*  S   UXSPTB  SOLUTION OF A SYSTEM OF LIN. EQUATIONS AFTER USING UXSPTF.
*  S   UXSPPB  SOLUTION OF A SYSTEM OF LIN. EQUATIONS WITH A SSOR PREC.
*  S   UXVSET  A SCALAR IS SET TO ALL THE ELEMENTS OF A VECTOR.
*  S   UXVDIR  VECTOR AUGMENTED BY THE SCALED VECTOR.
*  S   UXSSMD  MULT. OF A MATRIX BY A VECTOR AND ADDITION OF A SCALED VECTOR.
*  RF  UXVNOR  EUCLIDEAN NORM OF A VECTOR.
*  RF  UXVDOT  DOT PRODUCT OF TWO VECTORS.
*
      SUBROUTINE UXSSCG(N,M,G,IG,JG,A,MY,Z,OMEGA,X,XI,WD,WA,WR,WT,WP,WQ,
     &                  IW,EPS6,ETA2,MOS2,MOS4,NDEC,NMV,MCG,INF)
      INTEGER N,M,IG(N+1),JG(M),IW(N+1),MOS2,MOS4,NDEC,NMV,MCG,
     & NOS2,INF,I
      REAL*8 G(2*M),A(N),Z(N),X(N),MY,XI,OMEGA,EPS6,ETA2,
     & WD(N),WA(N),WR(N),WT(N),WP(N),WQ(N),
     & ALFA,BETA,GAMMA,DELTA,RHO,TAU,PI,PSI,ALF,SIG,
     & NORMZ,NORMR,UXVDOT,UXVNOR,TOL1,TOL2,
     & EPSZR,ZERO,ONE,WDMIN,WDMAX
      PARAMETER (EPSZR=1.0D-16,ZERO=0.0D 0,ONE=1.0D 0,
     &                          WDMIN=1.0D-3,WDMAX=1.0D 3)
C
      ALF=ETA2
      NOS2=MOS2
      IF (MCG.LE.0) MCG=N+3
C
      IF (MOS2.EQ.1) THEN
C
C     PRECONDITIONER C = D
C
      CALL UXSSDC(N,M,G,IG,WD)
      DO I=1,N
      IF (WD(I).LT.WDMIN) THEN
      WD(I)=ONE/WDMIN
      ELSE IF (WD(I).GT.WDMAX) THEN
      WD(I)=ONE/WDMAX
      ELSE
      WD(I)=ABS(ONE/WD(I))
      ENDIF
      ENDDO
      CALL UXVMUL(N,WD,A,WA,1)
      ELSE IF (MOS2.EQ.2) THEN
C
C     PRECONDITIONER C = L*D*TRANS(L)
C
      CALL UXVCOP(M,G,G(M+1))
      CALL UXSPTF(N,M,G(M+1),IG,JG,IW,INF,ALF,SIG)
      NDEC=NDEC+1
      CALL UXVCOP(N,A,WA)
      IF (INF.EQ.0.OR.MOS4.GT.0) THEN
      CALL UXSPTB(N,M,G(M+1),IG,JG,WA,0)
      ELSE
      NOS2=0
      ENDIF
      ELSE IF (MOS2.EQ.3) THEN
C
C     PRECONDITIONER C = (D+R)*INV(D)*(D+TRANS(R))
C
      CALL UXVCOP(N,A,WA)
      CALL UXSPPB(N,M,G,IG,JG,WA)
      ENDIF
C
C     INITIAL SETTINGS
C
      I=0
      CALL UXVSET(N,ZERO,X)
       XI=ZERO
       CALL UXVCOP(N,Z,WR)
       RHO=OMEGA
      CALL UXVSET(N,ZERO,WP)
       PI=ZERO
       BETA=ZERO
       DELTA=(UXVDOT(N,A,WA)-MY)
       IF (ABS(DELTA).LE.EPSZR**2) THEN
       INF=0
       GOTO 1
       ENDIF
      DELTA=ONE/DELTA
      NORMZ=UXVNOR(N,Z)
      NORMR=NORMZ
       TOL1=EPS6*NORMZ
       IF (TOL1.LT.EPSZR) TOL1=EPS6
       TOL2=EPS6*ABS(OMEGA)
       IF (TOL2.LT.EPSZR) TOL2=EPS6
C
C     CG ITERATIONS
C
      DO WHILE ((NORMR.GT.TOL1.OR.ABS(RHO).GT.TOL2).AND.I.LT.MCG)
       I=I+1
      TAU=DELTA*(UXVDOT(N,WA,WR)-RHO)
      CALL UXVDIR(N,-TAU,A,WR,WT)
      IF (NOS2.EQ.0) THEN
       CONTINUE
      ELSE IF (NOS2.EQ.1) THEN
      CALL UXVMUL(N,WD,WT,WT,1)
      ELSE IF (NOS2.EQ.2) THEN
      CALL UXSPTB(N,M,G(M+1),IG,JG,WT,0)
      ELSE IF (NOS2.EQ.3) THEN
      CALL UXSPPB(N,M,G,IG,JG,WT)
      ENDIF
      GAMMA=UXVDOT(N,WR,WT)+RHO*TAU
      BETA=BETA*GAMMA
      CALL UXVDIR(N,BETA,WP,WT,WP)
      PI=TAU+BETA*PI
      CALL UXSSMD(N,M,G,IG,JG,WP,PI,A,WQ)
      PSI=UXVDOT(N,A,WP)+MY*PI
      ALFA=UXVDOT(N,WP,WQ)+PI*PSI
       IF (ABS(ALFA).LE.EPSZR*ABS(GAMMA)) THEN
       INF=-I
       GOTO 1
       ENDIF
      ALFA=GAMMA/ALFA
      CALL UXVDIR(N,ALFA,WP,X,X)
      XI=XI+ALFA*PI
      CALL UXVDIR(N,-ALFA,WQ,WR,WR)
      RHO=RHO-ALFA*PSI
      BETA=ONE/GAMMA
      NORMR=UXVNOR(N,WR)
      ENDDO
      NMV=NMV+I
       INF=I
    1 RETURN
      END
